# -*- coding: utf-8 -*-

import pandas as pd
import numpy as np

Q_MAP = {
    'day': 250.0,
    'week': 52.0,
    'month': 12.0,
    'season': 4,
    'semi': 2,
    'year': 1
}


def norm_dist_draw(se_signal, n_draw):
    series = se_signal.copy()
    for i in range(n_draw):
        std = series.std()
        mean = series.mean()
        series[series < mean - 3 * std] = mean - 3 * std
        series[series > mean + 3 * std] = mean + 3 * std
    return series


def return_to_nav(nav_start_date, return_series):
    if return_series.empty:
        return pd.Series(index=[], data=[])
    nav_series = return_series.add(1.0).cumprod()
    nav_series.loc[nav_start_date] = 1.0
    return nav_series.sort_index()


def nav_to_return_patch(nav_series, daily_trading_date_list, trading_day_list):
    nav_series = nav_series.sort_index()

    if nav_series.empty:
        return nav_series
    fund_sub_daily_nav = nav_series.reindex(daily_trading_date_list).interpolate().dropna().sort_index()
    fund_sub_nav = fund_sub_daily_nav.reindex(trading_day_list).dropna().sort_index()

    return_series = fund_sub_nav.pct_change().dropna()
    daily_return_series = fund_sub_daily_nav.pct_change().dropna()
    return return_series, daily_return_series


def return_df_day_index_patch(fund_nav_dict, daily_trading_date_list, trading_day_list, is_norm=False):
    if fund_nav_dict is None:
        return pd.DataFrame(), pd.DataFrame()

    fund_return_dict = {}
    fund_daily_return_dict = {}

    for fund in fund_nav_dict.keys():
        fund_adj_nav = fund_nav_dict[fund]

        return_series, daily_return_series = nav_to_return_patch(
            fund_adj_nav, daily_trading_date_list, trading_day_list)

        if return_series.empty or daily_return_series.empty:
            continue

        if is_norm and len(return_series) > 1:
            return_series = norm_dist_draw(return_series, 1)
            daily_return_series = norm_dist_draw(daily_return_series, 1)

        fund_return_dict[fund] = return_series
        fund_daily_return_dict[fund] = daily_return_series

    return_df_day_index = pd.DataFrame.from_dict(fund_return_dict)
    daily_return_df_day_index = pd.DataFrame.from_dict(fund_daily_return_dict)
    return return_df_day_index, daily_return_df_day_index


def benchmark_return_series_computer(child_series, benchmark_return_df, quotedMargin_series=None, frequency='day'):
    q = Q_MAP[frequency]
    if benchmark_return_df.empty:
        return pd.Series(data=[])

    intersection_list = list(set(benchmark_return_df.columns) & set(child_series.index))
    benchmark_return_df = benchmark_return_df[intersection_list]
    child_series = child_series.ix[intersection_list]
    if quotedMargin_series is not None:
        quotedMargin_series = quotedMargin_series.ix[intersection_list]
        benchmark_return_series = benchmark_return_df.add(quotedMargin_series / q).dot(child_series)
    else:
        benchmark_return_series = benchmark_return_df.dot(child_series)
    return benchmark_return_series


def compute_geometric_annual_return_df(return_df, calc_q):
    t = return_df.shape[0]

    if t <= 0:
        return pd.DataFrame()

    holding_period_nav_series = return_df.fillna(0.).add(1.0).prod()
    tao = float(calc_q) / float(t)
    annual_return_series = holding_period_nav_series.pow(tao).subtract(1.0)
    return annual_return_series


def compute_arithmetic_annual_return_df(return_df, calc_q):
    mean_return = return_df.mean()
    annual_return_series = mean_return.multiply(calc_q)
    return annual_return_series


def compute_linking_annual_return_df(return_df, calc_q):
    t = return_df.shape[0]

    if t <= 0:
        return pd.DataFrame()

    holding_period_yield = return_df.fillna(0.).add(1.0).prod().subtract(1.0)
    tao = float(calc_q) / float(t)
    annual_return_series = holding_period_yield.multiply(tao)
    return annual_return_series


def compute_annual_return_df(return_df_day_index, calc_q, annual_type):
    if annual_type == 'arithmetic':
        annual_return = compute_arithmetic_annual_return_df(return_df=return_df_day_index, calc_q=calc_q)
    elif annual_type == 'linking':
        annual_return = compute_linking_annual_return_df(return_df=return_df_day_index, calc_q=calc_q)
    else:
        annual_return = compute_geometric_annual_return_df(return_df=return_df_day_index, calc_q=calc_q)
    return annual_return


def compute_geometric_annual_return(return_series, calc_q):
    t = len(return_series)

    if t <= 0:
        return np.NaN

    holding_period_nav = return_series.fillna(0.).add(1.0).prod()
    tao = float(calc_q) / float(t)
    return float((holding_period_nav ** tao) - 1.0)


def compute_arithmetic_annual_return(return_series, calc_q):
    mean_return = return_series.mean()
    return float(mean_return * calc_q)


def compute_linking_annual_return(return_series, calc_q):
    t = len(return_series)

    if t <= 0:
        return np.NaN

    holding_period_yield = return_series.fillna(0.).add(1.0).prod() - 1.0
    tao = float(calc_q) / float(t)
    return float(holding_period_yield * tao)


def compute_annual_return(return_series, calc_q, annual_type):
    if annual_type == 'arithmetic':
        annual_return = compute_arithmetic_annual_return(return_series=return_series, calc_q=calc_q)
    elif annual_type == 'linking':
        annual_return = compute_linking_annual_return(return_series=return_series, calc_q=calc_q)
    else:
        annual_return = compute_geometric_annual_return(return_series=return_series, calc_q=calc_q)
    return annual_return



import pandas as pd
import numpy as np
from datetime import datetime, timedelta
from hbshare.quant.CChen.cons import properties_stk_k, db, db_tables
from hbshare.quant.CChen.func import generate_table
from hbshare.quant.CChen.sql_cons import sql_quote
import hbshare as hbs
import pymysql

pymysql.install_as_MySQLdb()

page_size_con = 49999


def ch_stk_quote_to_local(db_path, sql_info, table='ch_stk_quote', page_size=None):
    if page_size is None:
        page_size = page_size_con

    from_table = db_tables['ch_stocks_daily_quote']

    generate_table(
        database='daily_data',
        table=table,
        generate_sql=sql_quote,
        sql_ip=sql_info['ip'],
        sql_user=sql_info['user'],
        sql_pass=sql_info['pass'],
        table_comment='from ' + from_table
    )
    print(table + ' generated')

    latest_date_in_db = pd.read_sql_query(
        'select distinct `TDATE` from ' + table + ' order by `TDATE` desc limit 1', db_path
    )
    if len(latest_date_in_db) > 0:
        latest_date = latest_date_in_db['TDATE'][0]
    else:
        latest_date = datetime(2000, 12, 31).date()

    print('\tLatest date in db: ' + str(latest_date))
    while 1:
        latest_date += timedelta(days=1)
        if latest_date > datetime.now().date():
            print('\t\t' + 'No more new quote data')
            return
        sql = (
                'select '
                'TDATE, '
                'SYMBOL, '
                'EXCHANGE, '
                'LCLOSE, '
                'TOPEN, '
                'TCLOSE, '
                'HIGH, '
                'LOW, '
                'VOTURNOVER, '
                'VATURNOVER, '
                'AVGPRICE, '
                'CHG, '
                'PCHG, '
                'PRANGE, '
                'MCAP, '
                'TCAP '
                'from ' + from_table
                + ' where  TDATE=' + latest_date.strftime('%Y%m%d')
        )

        data = hbs.db_data_query(db=db, sql=sql, page_size=page_size)
        df = pd.DataFrame(data['data'])
        if len(df) == 0:
            continue
        df['TDATE'] = pd.to_datetime(df['TDATE'].astype(str)).dt.date
        df = df.drop(columns=['ROW_ID']).rename(
            columns={
                'TOPEN': 'OPEN',
                'TCLOSE': 'CLOSE',
                'VOTURNOVER': 'VOLUME',
                'VATURNOVER': 'AMOUNT'
            }
        )
        df.to_sql(table, db_path, if_exists='append', index=False)
        print('\tNew data: ' + str(len(df)) + ', date: ' + latest_date.strftime('%Y-%m-%d'))


def trade_calendar(start_date=None, end_date=None, page_size=None):
    if start_date is None:
        start_date = datetime(2010, 1, 1).date()
    if end_date is None:
        end_date = datetime.now().date()
    if page_size is None:
        page_size = page_size_con

    sql = 'select distinct TDATE from ' + db_tables['ch_stocks_daily_quote'] \
          + ' where TDATE<=' + end_date.strftime('%Y%m%d') \
          + ' and TDATE>=' + start_date.strftime('%Y%m%d') \
          + ' order by TDATE'
    data = hbs.db_data_query(db=db, sql=sql, page_size=page_size)
    if data['pages'] > 1:
        for p in range(2, data['pages'] + 1):
            data['data'] = data['data'] + hbs.db_data_query(db=db, sql=sql, page_size=page_size, page_num=p)['data']

    return pd.to_datetime(pd.DataFrame(data['data'])['TDATE'], format='%Y%m%d').dt.date.tolist()


def daily_ret_by_vol(start_date=None, end_date=None, min_vol=10000, page_size=None):
    if start_date is None:
        start_date = datetime(2010, 1, 1).date()
    if end_date is None:
        end_date = datetime.now().date()
    if page_size is None:
        page_size = page_size_con

    sql = 'select ' + ', '.join(properties_stk_k) + ' from ' + db_tables['ch_stocks_daily_quote'] \
          + ' where TDATE<=' + end_date.strftime('%Y%m%d') \
          + ' and TDATE>=' + start_date.strftime('%Y%m%d') \
          + ' and VOTURNOVER>=' + str(min_vol) + ' order by TDATE, SYMBOL'

    data = hbs.db_data_query(db=db, sql=sql, page_size=page_size)
    if data['pages'] > 1:
        for p in range(2, data['pages'] + 1):
            data['data'] = data['data'] + hbs.db_data_query(db=db, sql=sql, page_size=page_size, page_num=p)['data']

    return data


def load_index(index_list, start_date=None, end_date=None, table=db_tables['index_daily_quote'], page_size=None):
    if start_date is None:
        start_date = datetime(2010, 1, 1).date()
    if end_date is None:
        end_date = datetime.now().date()
    if page_size is None:
        page_size = page_size_con

    if len(index_list) == 1:
        index_sql = '=\'' + index_list[0] + '\''
    elif len(index_list) > 1:
        index_sql = ' in ' + str(tuple(index_list))
    else:
        index_sql = '=\'000905\''

    sql = 'select ' \
          'JYRQ as tdate, ' \
          'ZQDM as code, ' \
          'QSPJ as pre_close, ' \
          'KPJG as open, ' \
          'SPJG as close, ' \
          'ZGJG as high, ' \
          'ZDJG as low, ' \
          'CJJS as amount,' \
          'ZDFD as PCHG from ' + table \
          + ' where JYRQ<=' + end_date.strftime('%Y%m%d') \
          + ' and JYRQ>=' + start_date.strftime('%Y%m%d') \
          + ' and ZQDM' + index_sql

    data = hbs.db_data_query(db=db, sql=sql, page_size=page_size)
    if data['pages'] > 1:
        for p in range(2, data['pages'] + 1):
            data['data'] = data['data'] + hbs.db_data_query(db=db, sql=sql, page_size=page_size, page_num=p)['data']

    return pd.DataFrame(data['data'])


def load_stk(start_date=None, end_date=None, table=db_tables['stocks_daily_quote'], page_size=None):
    if start_date is None:
        start_date = datetime(2010, 1, 1).date()
    if end_date is None:
        end_date = datetime.now().date()
    if page_size is None:
        page_size = page_size_con

    sql = 'select ' \
          'JYRQ, ' \
          'ZQDM as code, ' \
          'QSPJ as pre_close, ' \
          'KPJG as open, ' \
          'SPJG as close, ' \
          'ZGJG as high, ' \
          'ZDJG as low, ' \
          'HSBL as TURNOVER, ZDFD as PCHG from ' + table \
          + ' where JYRQ<=' + end_date.strftime('%Y%m%d') \
          + ' and JYRQ>=' + start_date.strftime('%Y%m%d') \
          + ' and (ZQDM like \'6%\' or ZQDM like \'3%\' or ZQDM like \'0%\') ' \
          + ' order by JYRQ, ZQDM'

    data = hbs.db_data_query(db=db, sql=sql, page_size=page_size)
    if data['pages'] > 1:
        for p in range(2, data['pages'] + 1):
            data['data'] = data['data'] + hbs.db_data_query(db=db, sql=sql, page_size=page_size, page_num=p)['data']

    return pd.DataFrame(data['data'])


def load_stk_local(sql_path, start_date=None, end_date=None, table='ch_stk_quote'):
    if start_date is None:
        start_date = datetime(2010, 1, 1).date()
    if end_date is None:
        end_date = datetime.now().date()

    data = pd.read_sql_query(
        'select * from ' + table
        + ' where TDATE<=' + end_date.strftime('%Y%m%d')
        + ' and TDATE>=' + start_date.strftime('%Y%m%d')
        + ' and (SYMBOL like \'6%\' or SYMBOL like \'3%\' or SYMBOL like \'0%\') '
        + ' order by TDATE, SYMBOL',
        sql_path
    )

    return data


def index_win(index_list, index_data, stk_data, freq='D'):

    cal = pd.DataFrame(pd.to_datetime(stk_data['TDATE'].drop_duplicates().sort_values()))
    cal = cal.set_index('TDATE', drop=False).resample(freq).last()
    cal = cal[~np.isnan(cal['TDATE'])]

    if len(cal) <= 1:
        print('No new data for frequency: ' + freq)
        return

    index_data['TDATE'] = pd.to_datetime(index_data['TDATE']).dt.date

    win_data = pd.DataFrame()

    for t in range(len(cal) - 1):
        t_date = cal.iloc[t][0].date()
        print(t_date)
        if t == 0:
            stk_data_t = stk_data[stk_data['TDATE'] <= t_date].reset_index(drop=True)
            index_data_t = index_data[index_data['TDATE'] <= t_date].reset_index(drop=True)
        else:
            stk_data_t = stk_data[
                np.array(stk_data['TDATE'] <= t_date)
                & np.array(stk_data['TDATE'] > cal.iloc[t - 1][0].date())
            ].reset_index(drop=True)
            index_data_t = index_data[
                np.array(index_data['TDATE'] <= t_date)
                & np.array(index_data['TDATE'] > cal.iloc[t - 1][0].date())
            ].reset_index(drop=True)

        stk_data_t_p = stk_data_t.pivot(index='TDATE', columns='CODE', values='PCHG')
        stk_data_t_p = (stk_data_t_p.fillna(0) + 100) / 100
        stk_data_t_p_r = (stk_data_t_p.prod() - 1) * 100

        index_data_t_p = index_data_t.pivot(index='TDATE', columns='CODE', values='PCHG')
        index_data_t_p = (index_data_t_p.fillna(0) + 100) / 100
        index_data_t_r = (index_data_t_p.prod() - 1) * 100

        stk_r_mean = stk_data_t_p_r.mean()
        stk_r_median = stk_data_t_p_r.median()
        stk_r_99 = stk_data_t_p_r.quantile(0.99)
        stk_r_95 = stk_data_t_p_r.quantile(0.95)
        stk_r_75 = stk_data_t_p_r.quantile(0.75)
        stk_r_25 = stk_data_t_p_r.quantile(0.25)
        stk_r_5 = stk_data_t_p_r.quantile(0.05)
        stk_r_1 = stk_data_t_p_r.quantile(0.01)

        stk_win_r = {}
        stk_win_n = {}
        for i in index_list:
            stk_win_r[i] = stk_data_t_p_r - index_data_t_r[i]
            stk_win_n[i + 'w'] = (stk_win_r[i] > 0).sum()

        result_dict = {
            'TDATE': t_date,
            'MEAN': stk_r_mean,
            'MEDIAN': stk_r_median,
            'Q99': stk_r_99,
            'Q95': stk_r_95,
            'Q75': stk_r_75,
            'Q25': stk_r_25,
            'Q5': stk_r_5,
            'Q1': stk_r_1,
            'STKN': stk_data_t_p_r.count()
        }
        result_dict.update(stk_win_n)
        result_dict.update(index_data_t_r)
        win_data = win_data.append(result_dict, sort=True, ignore_index=True)
        # win_data

    return win_data


if __name__ == '__main__':
    from db_cons import sql_write_path_work, sql_user_work
    # main = load_index(index_list='000905', start_date=datetime(2021, 10, 1))
    # main2 = load_stk(start_date=datetime(2021, 10, 1))

    # start_date = datetime(2021, 1, 1)
    # end_date = datetime.now()
    # index_list = [
    #     '000300',
    #     '000905'
    # ]
    # index_data = load_index(index_list=index_list, start_date=start_date, end_date=end_date)
    # stk_data = load_stk_local(start_date=start_date, end_date=end_date).rename(columns={'SYMBOL': 'CODE'})
    #
    # aa = index_win(
    #     index_data=index_data,
    #     stk_data=stk_data,
    #     index_list=index_list,
    #     freq='D'
    # )
    ch_stk_quote_to_local(db_path=sql_write_path_work['daily'], sql_info=sql_user_work)
    # print()





"""
国债收益率指标
"""
import pandas as pd
from datetime import datetime
from hbshare.quant.Kevin.asset_allocation.macro_index.util import create_table, delete_duplicate_records, WriteToDB
from WindPy import w

w.start()


class TreasuryYield:
    def __init__(self, start_date, end_date, is_increment=1):
        self.start_date = start_date
        self.end_date = end_date
        self.is_increment = is_increment
        self.table_name = 'mac_treasury_yield'

    def get_treasury_yield_data(self):
        """
        国债收益率数据：各期限国债收益率数据、美国十年期国债收益率数据
        """
        index_list = ['M1001645', 'M1001646', 'M1001647', 'M1001648', 'M1001650', 'M1001652', 'M1001654', 'M1001655',
                      'M1001656', 'M1001657', 'G0000891']
        name_dict = {"M1001645": "ytm_6m", "M1001646": "ytm_1y", "M1001647": "ytm_2y", "M1001648": "ytm_3y",
                     "M1001650": "ytm_5y", "M1001652": "ytm_7y", "M1001654": "ytm_10y", "M1001655": "ytm_15y",
                     "M1001656": "ytm_20y", "M1001657": "ytm_30y", "G0000891": "usa_ytm_10y"}
        res = w.edb(','.join(index_list), self.start_date, self.end_date)
        if res.ErrorCode != 0:
            data = pd.DataFrame()
            print("fetch treasury yield data error: start_date = {}, end_date = {}".format(self.start_date, self.end_date))
        else:
            if len(res.Data) == 1:
                data = pd.DataFrame(res.Data[0], index=res.Codes, columns=res.Times).T
            else:
                data = pd.DataFrame(res.Data, index=res.Codes, columns=res.Times).T
            data.index.name = 'trade_date'
            data.reset_index(inplace=True)
            data['trade_date'] = data['trade_date'].apply(lambda x: datetime.strftime(x, '%Y%m%d'))
            data.rename(columns=name_dict, inplace=True)

        return data

    def get_construct_result(self):
        if self.is_increment == 1:
            data = self.get_treasury_yield_data()
            sql_script = "delete from {} where trade_date in ({})".format(
                self.table_name, ','.join(data['trade_date'].tolist()))
            # delete first
            delete_duplicate_records(sql_script)
            # add new records
            WriteToDB().write_to_db(data, self.table_name)
        else:
            sql_script = """
                create table mac_treasury_yield(
                id int auto_increment primary key,
                trade_date date not null unique,
                ytm_6m decimal(6, 4),
                ytm_1y decimal(6, 4),
                ytm_2y decimal(6, 4),
                ytm_3y decimal(6, 4),
                ytm_5y decimal(6, 4),
                ytm_7y decimal(6, 4),
                ytm_10y decimal(6, 4),
                ytm_15y decimal(6, 4),
                ytm_20y decimal(6, 4),
                ytm_30y decimal(6, 4),
                usa_ytm_10y decimal(4, 2)) 
            """
            create_table(self.table_name, sql_script)
            data = self.get_treasury_yield_data()
            WriteToDB().write_to_db(data, self.table_name)


if __name__ == '__main__':
    TreasuryYield('2005-01-01', '2007-12-14', is_increment=1).get_construct_result()


import pandas as pd
from hbshare.fe.common.util.data_loader import NavAttributionLoader
import statsmodels.api as sm

fund_id = '510500'
fund_type = 'mutual'
start_date = '20210701'
end_date = '20211217'
factor_type = 'style'
benchmark_id = '000905'


data = NavAttributionLoader(fund_id, fund_type, start_date, end_date, factor_type, benchmark_id).load()
# preprocess
date_list = sorted(data['calendar_df']['calendarDate'].unique())
portfolio_nav_series = data['benchmark_series'].reindex(date_list).dropna().sort_index()
portfolio_nav_series = portfolio_nav_series / portfolio_nav_series.iloc[0]
portfolio_return_series = portfolio_nav_series.pct_change().dropna()
# factor
factor_nav_df = data['factor_data'] / data['factor_data'].iloc[0]
factor_return_df = factor_nav_df.pct_change().dropna().reindex(portfolio_return_series.index).fillna(0.)

model = sm.OLS(portfolio_return_series, sm.add_constant(factor_return_df)).fit()
df = pd.DataFrame(index=factor_return_df.columns.tolist(),
                  columns=['exposure', 'return_attr', 't_stat', 'p_value'])
df.loc[:, 'exposure'] = model.params.drop('const')
df.loc[:, 't_stat'] = model.tvalues.drop('const')
df['t_stat'] = df['t_stat'].round(4)
df.loc[:, 'p_value'] = model.pvalues.drop('const')
df['p_value'] = df['p_value'].round(4)
# cal_return
factor_cum_return = factor_return_df.fillna(0.).add(1.0).prod().subtract(1.0)
perf_attr = df['exposure'].reindex(factor_cum_return.index).multiply(factor_cum_return)
df.loc[:, 'return_attr'] = perf_attr
df.round(4).to_clipboard()


# -*- coding: utf-8 -*-

import pandas as pd
import pymysql
import hbshare as hbs
import warnings
warnings.filterwarnings('ignore', category=pymysql.Warning)


class Loader:
    def __init__(self):
        pass

    def get_df(self, sql, db, page_size=2000):
        data = hbs.db_data_query(db, sql, page_size=page_size, timeout=120)
        pages = data['pages']
        data = pd.DataFrame(data['data'])
        if pages > 1:
            for page in range(2, pages + 1):
                temp_data = hbs.db_data_query(db, sql, page_size=page_size, page_num=page, timeout=120)
                data = pd.concat([data, pd.DataFrame(temp_data['data'])], axis=0)
        return data

    def read_cal(self, start, end):
        sql = "SELECT JYRQ, SFJJ, SFZM, SFYM FROM funddb.JYRL WHERE JYRQ>={0} and JYRQ<={1}".format(start, end)
        df = self.get_df(sql, db='readonly')
        return df

    def read_mutual_index_daily_k_given_indexs(self, indexs, start, end):
        indexs = "'" + "','".join(indexs) + "'"
        sql = "SELECT zsdm AS INDEX_CODE, jzrq AS TRADE_DATE, spjg AS CLOSE_INDEX FROM st_fund.t_st_gm_clzs WHERE m_opt_type <> '03' AND zsdm in ({0}) AND jzrq >= {1} AND jzrq <= {2}".format(indexs, start, end)
        df = self.get_df(sql, db='funduser', page_size=200000)
        return df

    def read_private_index_daily_k_given_indexs(self, indexs, start, end):
        indexs = "'" + "','".join(indexs) + "'"
        sql = "SELECT zsdm AS INDEX_CODE, tjyf AS TRADE_MONTH, spjg AS CLOSE_INDEX FROM st_hedge.t_st_sm_hmzs WHERE m_opt_type <> '03' AND zsdm in ({0}) AND tjyf >= {1} AND tjyf <= {2}".format(indexs, start, end)
        df = self.get_df(sql, db='highuser', page_size=200000)
        return df

    def read_market_index_daily_k_given_indexs(self, indexs, start, end):
        indexs = "'" + "','".join(indexs) + "'"
        sql = "SELECT zqdm AS INDEX_CODE, jyrq AS TRADE_DATE, spjg AS CLOSE_INDEX FROM st_market.t_st_zs_hqql WHERE m_opt_type <> '03' AND zqdm in ({0}) AND jyrq >= {1} AND jyrq <= {2}".format(indexs, start, end)
        df = self.get_df(sql, db='alluser', page_size=200000)
        return df

    def read_mutual_fund_cumret_given_codes(self, codes, start, end):
        codes = "'" + "','".join(codes) + "'"
        sql = "SELECT a.jjdm AS FUND_CODE, b.jzrq AS TRADE_DATE, b.hbcl AS CUM_RET FROM funddb.jjxx1 a, funddb.jjhb b WHERE a.cpfl = '2' AND a.jjdm = b.jjdm AND a.jjzt not in ('3', 'c') AND a.m_opt_type <> '03' AND a.jjdm in ({0}) AND b.jzrq >= {1} AND b.jzrq <= {2}".format(codes, start, end)
        df = self.get_df(sql, db='readonly', page_size=200000)
        return df

    def read_private_fund_adj_nav_given_codes(self, codes, start, end):
        codes = "'" + "','".join(codes) + "'"
        sql = "SELECT a.jjdm AS FUND_CODE, b.jzrq AS TRADE_DATE, b.fqdwjz AS ADJ_NAV FROM st_hedge.t_st_jjxx a, st_hedge.t_st_rhb b WHERE a.cpfl = '4' AND a.jjdm = b.jjdm AND a.jjzt not in ('3') AND a.m_opt_type <> '03' AND a.jjdm in ({0}) AND b.jzrq >= {1} AND b.jzrq <= {2}".format(codes, start, end)
        df = self.get_df(sql, db='highuser', page_size=200000)
        return df

    def read_market_index_market_value_given_indexs(self, indexs, start, end):
        indexs = "'" + "','".join(indexs) + "'"
        sql = "SELECT zqdm AS INDEX_CODE, jyrq AS TRADE_DATE, zsz AS MARKET_VALUE FROM st_market.t_st_zs_hqql WHERE m_opt_type <> '03' AND zqdm in ({0}) AND jyrq >= {1} AND jyrq <= {2}".format(indexs, start, end)
        df = self.get_df(sql, db='alluser', page_size=200000)
        return df

    def read_mutual_fund_market_value_given_codes(self, codes, start, end):
        codes = "'" + "','".join(codes) + "'"
        sql = "SELECT jjdm AS FUND_CODE, jsrq AS REPORT_DATE, jjzzc AS MARKET_VALUE FROM st_fund.t_st_gm_zcpz WHERE m_opt_type <> '03' AND jjdm in ({0}) AND jsrq >= {1} AND jsrq <= {2}".format(codes, start, end)
        df = self.get_df(sql, db='funduser', page_size=200000)
        return df


"""
本地净值对比程序
"""
import pandas as pd
import numpy as np
from datetime import datetime
import hbshare as hbs
from sqlalchemy import create_engine
from hbshare.quant.Kevin.rm_associated.config import engine_params
from hbshare.fe.common.util.data_loader import get_trading_day_list
from Arbitrage_backtest import cal_annual_return, cal_annual_volatility, cal_sharpe_ratio, cal_max_drawdown
import plotly
from plotly.offline import plot as plot_ly
import plotly.graph_objs as go
import plotly.figure_factory as ff

plotly.offline.init_notebook_mode(connected=True)


class Neutral2Long:
    """
    中性净值剥离
    """
    def __init__(self, neutral_series, fund_name, hedge_ratio=1):
        self.neutral_series = neutral_series
        self.fund_name = fund_name
        self.hedge_ratio = hedge_ratio
        self._load_data()

    def _load_data(self):
        start_date, end_date = self.neutral_series.index[0], self.neutral_series.index[-1]

        sql_script = "SELECT JYRQ, SFJJ, SFZM, SFYM FROM funddb.JYRL WHERE JYRQ >= {} and JYRQ <= {}".format(
            start_date, end_date)
        res = hbs.db_data_query('readonly', sql_script, page_size=5000)
        df = pd.DataFrame(res['data']).rename(
            columns={"JYRQ": 'calendarDate', "SFJJ": 'isOpen',
                     "SFZM": "isWeekEnd", "SFYM": "isMonthEnd"}).sort_values(by='calendarDate')
        df['isOpen'] = df['isOpen'].astype(int).replace({0: 1, 1: 0})
        df['isWeekEnd'] = df['isWeekEnd'].fillna(0).astype(int)
        df['isMonthEnd'] = df['isMonthEnd'].fillna(0).astype(int)
        trading_day_list = df[df['isWeekEnd'] == 1]['calendarDate'].tolist()

        sql_script = "SELECT * FROM mac_share_price_index_futures where trade_date >= {} and trade_date <= {}".format(
            start_date, end_date)
        engine = create_engine(engine_params)
        data = pd.read_sql(sql_script, engine)
        data['trade_date'] = data['trade_date'].apply(lambda x: datetime.strftime(x, '%Y%m%d'))

        self.data = data[['trade_date', 'ZZ500', 'IC00', 'IC01']].set_index('trade_date').reindex(trading_day_list)

    @staticmethod
    def week_of_month(dt):
        """ Returns the week of the month for the specified date.
        """
        first_day = dt.replace(day=1)
        dom = dt.day
        adjusted_dom = dom + first_day.weekday()
        return int(np.ceil(adjusted_dom / 7.0))

    def get_construct_result(self):
        data = self.data.copy()
        df = pd.merge(self.neutral_series.to_frame('neutral_nav'), data, left_index=True, right_index=True).sort_index()
        df['index_return'] = df['ZZ500'].pct_change().fillna(0.)
        df['future_return'] = df['IC00'].pct_change().fillna(0.)
        df['trade_date'] = df.index
        df['week'] = df['trade_date'].apply(lambda x: self.week_of_month(datetime.strptime(x, '%Y%m%d')))
        df.loc[df['week'] == 3, 'roll_return'] = df['IC01'] / df['IC00'] - 1
        df['roll_return'] = df['roll_return'].fillna(0.)
        df['单日对冲成本'] = (df['index_return'] - df['future_return'] + df['roll_return']) * self.hedge_ratio
        df['nav_对冲成本'] = (1 + df['单日对冲成本']).cumprod()
        df['对冲前alpha'] = (df['neutral_nav'] - df['nav_对冲成本']) / self.hedge_ratio + 1
        # plus beta
        df['return'] = df['index_return'] + df['对冲前alpha'].pct_change().fillna(0.)
        df[self.fund_name] = (1 + df['return']).cumprod()

        return df[[self.fund_name]]

# 概率一号 vs 概率高频一号
from hbshare.quant.Kevin.quant_room.MyUtil.data_loader import get_fund_nav_from_sql
nav_series1 = get_fund_nav_from_sql('20211231', '20220422', {"概率一号": "SNM976"})
alpha_series1 = Neutral2Long(nav_series1['概率一号'], "概率一号", 0.8).get_construct_result()
# nav_series2 = get_fund_nav_from_sql('20211231', '20220422', {"概率高频一号": "SNM980"})
# alpha_series2 = Neutral2Long(nav_series2['概率高频一号'], "概率高频一号", 0.8).get_construct_result()

# from hbshare.quant.Kevin.quant_room.MyUtil.data_loader import get_fund_nav_from_sql
# nav_series = get_fund_nav_from_sql('20201220', '20220311', {"明汯中性一号": "ST7452"})
# a = Neutral2Long(nav_series['明汯中性一号'], "明汯中性1号", 0.8).get_construct_result()

# 伯兄
# local_nav = pd.read_excel("D:\\研究基地\\A-机器学习类\\伯兄\\净值文件\\伯兄卢比孔-多头剥离.xlsx", sheet_name=0, index_col=0)
# local_nav['tradeDate'] = local_nav.index
# local_nav['tradeDate'] = local_nav['tradeDate'].apply(lambda x: datetime.strftime(x, '%Y%m%d'))
# local_nav.rename(columns={"对冲前alpha（100资金）": "伯兄500指增"}, inplace=True)
# local_nav = local_nav.set_index('tradeDate')['伯兄500指增'].sort_index()
# local_nav = Neutral2Long(local_nav, '伯兄500指增').get_construct_result()


# 概率
# nav_pieces1 = pd.read_excel('D:\\研究基地\\A-机器学习类\\概率\\概率业绩数据_0903.xlsx', sheet_name=5)
# nav_pieces1.rename(columns={"日期": "trade_date", "估算超额收益": "estimate_return"}, inplace=True)
# nav_pieces1['trade_date'] = nav_pieces1['trade_date'].apply(lambda x: datetime.strftime(x, "%Y%m%d"))
# nav_pieces1 = Neutral2Long(
#     nav_pieces1.set_index('trade_date')['累计净值'], '概率500指增', hedge_ratio=0.8).get_construct_result()
#
# sql_script_out = "SELECT a.jjdm fund_id, b.jzrq TRADEDATE, b.fqdwjz as ADJ_NAV from " \
#              "st_hedge.t_st_jjxx a, st_hedge.t_st_rhb b where a.cpfl = '4' and a.jjdm = b.jjdm " \
#              "and a.jjzt not in ('3') " \
#              "and a.jjdm = '{}' and b.jzrq >= {} and b.jzrq <= {} " \
#              "order by b.jzrq".format('SQT076', '20210625', '20211217')
# res_out = hbs.db_data_query("highuser", sql_script_out, page_size=5000)
# online_nav = pd.DataFrame(res_out['data']).set_index('TRADEDATE')['ADJ_NAV']
# local_nav = pd.concat([nav_pieces1, online_nav], axis=1)
# local_nav['return'] = local_nav['ADJ_NAV'].pct_change()
# local_nav['tmp'] = local_nav['概率500指增']
# start_point = local_nav['return'].first_valid_index()
# for k in range(local_nav.index.tolist().index(start_point), len(local_nav)):
#     local_nav.loc[local_nav.index.tolist()[k], 'tmp'] = local_nav.iloc[k - 1]['tmp'] * (1 + local_nav.iloc[k]['return'])
# local_nav = local_nav[['tmp']].rename(columns={"tmp": "概率500指增"})


# 白鹭
# from hbshare.quant.Kevin.quant_room.MyUtil.data_loader import get_fund_nav_from_sql
# nv_series1 = get_fund_nav_from_sql('20200306', '20211231', {"白鹭alpha": "SJM897"})
# return_series1 = nv_series1.pct_change().fillna(0.)
# nv_series2 = get_fund_nav_from_sql('20211231', '20220318', {"白鹭alpha": "SQB109"})
# return_series2 = nv_series2.pct_change().dropna()
# return_series = pd.concat([return_series1, return_series2], axis=0).sort_index()
# local_nav = (1 + return_series).cumprod()

# 希格斯： 回测 + 实盘
# data1 = pd.read_excel('D:\\希格斯投资Alpha系列-实盘&回测超额表现20220415.xlsx', sheet_name=5)
# data1['日期'] = data1['日期'].apply(lambda x: datetime.strftime(x, '%Y%m%d'))
# data1 = data1.set_index('日期')['相对500每日超额']
# data1 = data1[data1.index < '20210819']
# data2 = pd.read_excel('D:\\希格斯投资Alpha系列-实盘&回测超额表现20220415.xlsx', sheet_name=4)
# data2['日期'] = data2['日期'].apply(lambda x: datetime.strftime(x, '%Y%m%d'))
# data2['相对500每日超额'] = np.where(
#     data2['水手2号相对500每日超额'].isnull(), data2['水手3号相对500每日超额'], data2['水手2号相对500每日超额']) / 100.
# data2 = data2.set_index('日期')['相对500每日超额']
# data2 = data2[data2.index >= '20210819']
# excess_series = pd.concat([data1, data2])
#
# trading_day_list = get_trading_day_list('20180214', '20220415')
# benchmark_id = '000905'
# sql_script = "SELECT JYRQ as TRADEDATE, ZQMC as INDEXNAME, SPJG as TCLOSE from funddb.ZSJY WHERE ZQDM = '{}' " \
#              "and JYRQ >= {} and JYRQ <= {}".format(benchmark_id, '20180214', '20220415')
# res = hbs.db_data_query('readonly', sql_script, page_size=2000)
# data = pd.DataFrame(res['data']).rename(columns={"TCLOSE": "benchmark"}).set_index('TRADEDATE')[['benchmark']]
# data = data.reindex(trading_day_list)
# data = data.pct_change().dropna()['benchmark']
# excess_series = excess_series.reindex(data.index)
# return_series = excess_series + data
# trading_day_list = get_trading_day_list('20180214', '20220415', frequency='week')
# local_nav = (1 + return_series).cumprod().reindex(trading_day_list).to_frame('希格斯水手2号')


# 伯兄建康
# from hbshare.quant.Kevin.quant_room.MyUtil.data_loader import get_fund_nav_from_sql
# local_nav = get_fund_nav_from_sql('20210730', '20220325', {"伯兄建康": "SQT564"})

# nv_series = pd.read_excel('D:\\研究基地\\D-量价类\\翰荣\\翰荣指数增强策略净值分析2022.04.08.xlsx', sheet_name=0, header=1)
# nv_series['日期'] = nv_series['日期'].apply(lambda x: datetime.strftime(x, '%Y%m%d'))
# nv_series.rename(columns={"策略净值": "翰荣500指增"}, inplace=True)
# local_nav = nv_series.set_index('日期')['翰荣500指增']

# from hbshare.quant.Kevin.quant_room.MyUtil.data_loader import get_fund_nav_from_sql
# local_nav = get_fund_nav_from_sql('20210305', '20220415', {"星阔广厦": "SNU706"})

from hbshare.quant.Kevin.quant_room.MyUtil.data_loader import get_fund_nav_from_sql
local_nav = get_fund_nav_from_sql('20210618', '20220729', {"概率": "SQQ803"})

compare_dict = {"衍复": "SJH866",
                "启林": "SGY379",
                "因诺": "SGX346",
                "天演": "P20830",
                "诚奇": "SLS817",
                "明汯": "SS5789",
                # "灵均进取1号": "SW3470",
                "凡二": "SJM016",
                "星阔": "SNU706",
                "世纪前沿": "SGP682",
                "赫富": "SEP463",
                "九坤": "ST9804",
                "乾象": "P48470",
                # "白鹭": "SQB109",
                # "衍合500指数增强1号": "SLG867",
                # "朋锦永宁": "SGN254",
                # "弈倍虎鲸": "SY0096"
                }


# compare_dict = {"衍复": "SJM688",
#                 "明汯": "SGG585",
#                 "九坤": "SCP381",
#                 # "QL": "SGY044",
#                 "星阔": "SQF225"}


class LocalNavComparer:
    def __init__(self, start_date, end_date):
        self.start_date = start_date
        self.end_date = end_date
        self._load_data()

    def _load_data(self):
        # calendar
        trading_day_list = get_trading_day_list(self.start_date, self.end_date, frequency='week')
        # nav_df
        nav_list = []
        for fund_name, fund_id in compare_dict.items():
            sql_script = "SELECT a.jjdm fund_id, b.jzrq TRADEDATE, b.fqdwjz as ADJ_NAV from " \
                         "st_hedge.t_st_jjxx a, st_hedge.t_st_rhb b where a.cpfl = '4' and a.jjdm = b.jjdm " \
                         "and a.jjzt not in ('3') " \
                         "and a.jjdm = '{}' and b.jzrq >= {} and b.jzrq <= {} " \
                         "order by b.jzrq".format(fund_id, self.start_date, self.end_date)
            res = hbs.db_data_query("highuser", sql_script, page_size=5000)
            data = pd.DataFrame(res['data']).set_index('TRADEDATE')['ADJ_NAV']
            data.name = fund_name
            nav_list.append(data)
        nav_df = pd.concat(nav_list, axis=1).sort_index().reindex(trading_day_list).fillna(method='ffill', limit=1)
        nav_df = pd.merge(nav_df, local_nav, left_index=True, right_index=True, how='left').fillna(
            method='ffill', limit=1)
        nav_df = nav_df / nav_df.iloc[0]

        nav_df = nav_df.dropna(axis=1)

        # benchmark
        benchmark_id = '000905'
        sql_script = "SELECT JYRQ as TRADEDATE, ZQMC as INDEXNAME, SPJG as TCLOSE from funddb.ZSJY WHERE ZQDM = '{}' " \
                     "and JYRQ >= {} and JYRQ <= {}".format(benchmark_id, self.start_date, self.end_date)
        res = hbs.db_data_query('readonly', sql_script)
        data = pd.DataFrame(res['data']).rename(columns={"TCLOSE": "benchmark"}).set_index('TRADEDATE')[['benchmark']]
        data = data.reindex(trading_day_list)

        assert (nav_df.shape[0] == data.shape[0])

        excess_return = nav_df.pct_change().fillna(0.).sub(data.pct_change()['benchmark'].squeeze().fillna(0.), axis=0)

        self.excess_df = (1 + excess_return.fillna(0.)).cumprod()

    @staticmethod
    def plotly_line(df, title_text, sava_path, figsize=(1200, 500)):
        fig_width, fig_height = figsize
        data = []
        for col in df.columns[:-1]:
            trace = go.Scatter(
                x=df.index.tolist(),
                y=df[col],
                name=col,
                mode="lines"
            )
            data.append(trace)

        col = df.columns[-1]
        trace = go.Scatter(
                x=df.index.tolist(),
                y=df[col],
                name=col,
                mode="lines",
                line=dict(dash='dot', width=3))
        data.append(trace)

        date_list = df.index.tolist()
        n = int(len(date_list) / 12)
        tick_vals = [i for i in range(0, len(df), n)]
        tick_text = [date_list[i] for i in range(0, len(df), n)]

        layout = go.Layout(
            title=dict(text=title_text),
            autosize=False, width=fig_width, height=fig_height,
            yaxis=dict(tickfont=dict(size=12), showgrid=True),
            xaxis=dict(showgrid=True, tickvals=tick_vals, ticktext=tick_text),
            template='plotly_white'
        )
        fig = go.Figure(data=data, layout=layout)

        plot_ly(fig, filename=sava_path, auto_open=False)

    def get_construct_result(self):
        # plot
        self.plotly_line(
            self.excess_df, "500指增产品对比", "D:\\量化产品跟踪\\新增净值走势对比\\净值曲线.html", figsize=(1500, 800))
        # calculate
        nav_df = self.excess_df.copy()
        portfolio_index_df = pd.DataFrame(
            index=nav_df.columns, columns=['超额年化收益', '超额年化波动', '最大回撤', 'Sharpe', '胜率', '平均损益比'])
        portfolio_index_df.loc[:, '超额年化收益'] = nav_df.pct_change().dropna(how='all').apply(cal_annual_return, axis=0)
        portfolio_index_df.loc[:, '超额年化波动'] = \
            nav_df.pct_change().dropna(how='all').apply(cal_annual_volatility, axis=0)
        portfolio_index_df.loc[:, '最大回撤'] = nav_df.apply(cal_max_drawdown, axis=0)
        portfolio_index_df.loc[:, 'Sharpe'] = \
            nav_df.pct_change().dropna(how='all').apply(lambda x: cal_sharpe_ratio(x, 0.015), axis=0)
        portfolio_index_df.loc[:, '胜率'] = \
            nav_df.pct_change().dropna(how='all').apply(lambda x: x.gt(0).sum() / len(x), axis=0)
        portfolio_index_df.loc[:, '平均损益比'] = \
            nav_df.pct_change().dropna(how='all').apply(lambda x: x[x > 0].mean() / x[x < 0].abs().mean(), axis=0)
        portfolio_index_df.index.name = '产品名称'
        portfolio_index_df.reset_index(inplace=True)
        # 格式处理
        portfolio_index_df['超额年化收益'] = portfolio_index_df['超额年化收益'].apply(lambda x: format(x, '.2%'))
        portfolio_index_df['超额年化波动'] = portfolio_index_df['超额年化波动'].apply(lambda x: format(x, '.2%'))
        portfolio_index_df['最大回撤'] = portfolio_index_df['最大回撤'].apply(lambda x: format(x, '.2%'))
        portfolio_index_df['Sharpe'] = portfolio_index_df['Sharpe'].round(2)
        portfolio_index_df['胜率'] = portfolio_index_df['胜率'].apply(lambda x: format(x, '.1%'))
        portfolio_index_df['平均损益比'] = portfolio_index_df['平均损益比'].round(2)

        fig = ff.create_table(portfolio_index_df)
        fig.layout.autosize = False
        fig.layout.width = 1000
        fig.layout.height = 400

        plot_ly(fig, filename="D:\\量化产品跟踪\\新增净值走势对比\\收益指标统计.html", auto_open=False)


if __name__ == '__main__':
    LocalNavComparer("20210618", "20220729").get_construct_result()


"""
日度净值归因模块
"""
import pandas as pd
import numpy as np
import hbshare as hbs
from datetime import datetime
from hbshare.quant.Kevin.quant_room.MyUtil.data_loader import get_trading_day_list, get_fund_nav_from_sql
from hbshare.quant.Kevin.rm_associated.config import style_names
import statsmodels.api as sm


class AlphaNavAttribution:
    def __init__(self, start_date, end_date, token):
        self.start_date = start_date
        self.end_date = end_date
        self.token = token
        self._load_data()

    @staticmethod
    def nav_transform(nav_df):
        nav_df['div'] = nav_df['dividend'].diff()
        nav_df['div'] = nav_df['div'].fillna(nav_df['dividend'].iloc[0])
        nav_df['adj_factor'] = nav_df['jjjz'].shift(1) / (nav_df['jjjz'].shift(1) - nav_df['div'])
        nav_df['adj_factor'] = nav_df['adj_factor'].fillna(1.).cumprod()
        nav_df['adj_nav'] = nav_df['adj_factor'] * nav_df['jjjz']

        return nav_df['adj_nav']

    @staticmethod
    def nav_check(adj_nav, fund_id):
        start_date, end_date = adj_nav.index[0], adj_nav.index[-1]
        nv_series = get_fund_nav_from_sql(start_date, end_date, {"actual": fund_id})
        trading_day_list = get_trading_day_list(start_date, end_date, "week")
        compare_df = adj_nav.to_frame('calculate').merge(
            nv_series, left_index=True, right_index=True).reindex(trading_day_list)

        return compare_df

    def nav_process(self, all_data, fund_name, fof_name):
        selected_data = all_data[all_data['jjmc'] == fund_name].dropna(subset=['jjjz', 'ljjz'])
        selected_data = selected_data[selected_data['khmc'] == fof_name].sort_values(by='jzrq')  # longest period
        fund_id = selected_data['jjdm'].unique()[0]
        selected_data = selected_data.set_index('jzrq')[['jjjz', 'ljjz']]
        start_date, end_date = selected_data.index[0], selected_data.index[-1]
        date_list = get_trading_day_list(start_date, end_date)
        origin_nav = selected_data.reindex(date_list).astype(float)
        origin_nav['dividend'] = (origin_nav['ljjz'] - origin_nav['jjjz']).round(4)
        adj_nav = self.nav_transform(origin_nav).sort_index()
        # test
        test_df = self.nav_check(adj_nav, fund_id)
        test_df = test_df.pct_change().dropna()
        test_df['delta'] = (test_df['calculate'] - test_df['actual']).abs()
        print("【{}】在区间内日度收益的总误差为: {}".format(fund_name, test_df['delta'].sum()))

        return adj_nav

    def _load_data(self):
        # 所有数据
        all_data = hbs.commonQuery('FOF_ZJJLIST', startDate=self.start_date, endDate=self.end_date,
                                   sorts=[{"field": "jzrq", "sort": "desc"}],
                                   access_token=self.token,
                                   fields=['jjdm', 'jjmc', 'khdm', 'khmc', 'jjjz', 'jzrq', 'ljjz'])
        # 乾象、伯兄、启林、因诺、明汯(1000)、星阔(500 + 全市场)、概率(500 + 1000)、诚奇
        nav_dict = {
            "500指增": {},
            "1000指增": {},
            "全市场": {}
        }
        # special treat
        all_data.replace({'乾象中证500指数增强1号私募证券投资基金B类': '乾象中证500指数增强1号私募证券投资基金B',
                          '明汯量化中小盘增强1号私募证券投资基金A': '明汯量化中小盘增强1号私募证券投资基金'}, inplace=True)

        # 乾象(特殊处理)
        tmp = all_data[(all_data['jjmc'].str.contains('乾象')) & (all_data['jjmc'].str.contains('500指数增强'))]
        tmp = tmp.dropna(subset=['jjjz', 'ljjz'])
        tmp = tmp[tmp['khmc'] == '新方程大类配置基金'].sort_values(by='jzrq')  # 最长的净值
        tmp = tmp.set_index('jzrq')[['jjjz', 'ljjz']]
        trading_day_list = get_trading_day_list(tmp.index[0], tmp.index[-1])
        tmp = tmp.reindex(trading_day_list)
        tmp.loc['20220429'] = ['0.8434', '1.1350']  # 手动填充
        tmp = tmp.astype(float)
        tmp['dividend'] = (tmp['ljjz'] - tmp['jjjz']).round(4)
        adj_nav = self.nav_transform(tmp).sort_index()
        test_df = self.nav_check(adj_nav, 'P48470').pct_change().dropna()
        test_df['delta'] = (test_df['calculate'] - test_df['actual']).abs()
        print("区间内日度收益的总误差为: {}".format(test_df['delta'].sum()))
        nav_dict['500指增']['乾象'] = adj_nav
        # 伯兄
        adj_nav = self.nav_process(all_data, '伯兄熙宁中证500指数增强私募证券投资基金', '新方程量化中小盘精选二号私募证券投资基金')
        nav_dict['500指增']['伯兄'] = adj_nav
        # 启林(500 + 1000)
        adj_nav = self.nav_process(all_data, '启林中证500指数增强1号私募证券投资基金', '新方程量化中小盘精选私募证券投资基金')
        nav_dict['500指增']['启林'] = adj_nav
        adj_nav = self.nav_process(all_data, '启林玉成一号私募证券投资基金B', '新方程量化中小盘精选二号私募证券投资基金')
        nav_dict['1000指增']['启林'] = adj_nav.loc["20211231":]
        # 因诺
        adj_nav = self.nav_process(all_data, '因诺聚配中证500指数增强私募证券投资基金', '新方程量化中小盘精选私募证券投资基金')
        adj_nav.loc['20211126'] = 2.2350
        nav_dict['500指增']['因诺'] = adj_nav
        # 明汯(500 + 1000)
        adj_nav = self.nav_process(all_data, '明汯价值成长1期3号私募投资基金', '新方程大类配置基金')
        nav_dict['500指增']['明汯'] = adj_nav
        adj_nav = self.nav_process(all_data, '明汯量化中小盘增强1号私募证券投资基金', '新方程量化中小盘精选私募证券投资基金')
        adj_nav.loc['20211126'] = 2.5951
        nav_dict['1000指增']['明汯'] = adj_nav
        # 星阔
        adj_nav = self.nav_process(all_data, '星阔广厦1号中证500指数增强私募证券投资基金', '新方程量化中小盘精选私募证券投资基金')
        nav_dict['500指增']['星阔'] = adj_nav
        adj_nav = self.nav_process(all_data, '星阔山海6号股票优选私募证券投资基金', '新方程对冲精选H1号基金')
        nav_dict['全市场']['星阔'] = adj_nav
        # 概率(500 + 1000)
        adj_nav = self.nav_process(all_data, '概率500指增2号私募证券投资基金', '新方程量化中小盘精选二号私募证券投资基金')
        nav_dict['500指增']['概率'] = adj_nav
        adj_nav = self.nav_process(all_data, '概率1000指增1号私募证券投资基金', '新方程量化中小盘精选二号私募证券投资基金')
        nav_dict['1000指增']['概率'] = adj_nav
        # 诚奇
        adj_nav = self.nav_process(all_data, '诚奇睿盈500指数增强私募证券投资基金A', '新方程好买臻选3号私募证券投资母基金')
        nav_dict['500指增']['诚奇'] = adj_nav
        # 幻方
        adj_nav = self.nav_process(all_data, '幻方500指数增强欣享6号私募证券投资基金', '新方程量化中小盘精选私募证券投资基金')
        nav_dict['500指增']['幻方'] = adj_nav
        # 天演(500 + 全市场)
        adj_nav = self.nav_process(all_data, '天演中证500指数增强证券投资基金', '新方程量化中小盘精选私募证券投资基金')
        nav_dict['500指增']['天演'] = adj_nav.loc['20210625':]
        adj_nav = self.nav_process(all_data, '天演赛能证券投资基金', '新方程大类配置基金')
        nav_dict['全市场']['天演'] = adj_nav
        # 衍复(500 + 1000, 500特殊处理)
        tmp = all_data[(all_data['jjmc'].str.contains('衍复')) & (all_data['jjmc'].str.contains('指增三号'))]
        tmp = tmp.dropna(subset=['jjjz', 'ljjz'])
        tmp = tmp[tmp['khmc'] == '新方程量化中小盘精选私募证券投资基金'].sort_values(by='jzrq')  # 最长的净值
        tmp = tmp.set_index('jzrq')[['jjjz', 'ljjz']]
        trading_day_list = get_trading_day_list(tmp.index[0], tmp.index[-1])
        tmp = tmp.reindex(trading_day_list)
        tmp.loc['20211126'] = ['1.4791', '1.8191']
        tmp.loc['20220113'] = ['1.3665', '1.8065']  # 手动填充
        tmp = tmp.astype(float).reindex(trading_day_list)
        tmp['dividend'] = (tmp['ljjz'] - tmp['jjjz']).round(4)
        adj_nav = self.nav_transform(tmp).sort_index()
        test_df = self.nav_check(adj_nav, 'SJH866').pct_change().dropna()
        test_df['delta'] = (test_df['calculate'] - test_df['actual']).abs()
        print("衍复500区间内日度收益的总误差为: {}".format(test_df['delta'].sum()))
        nav_dict['500指增']['衍复'] = adj_nav

        adj_nav = self.nav_process(all_data, '衍复1000指增一号私募证券投资基金', '新方程量化中小盘精选私募证券投资基金')
        nav_dict['1000指增']['衍复'] = adj_nav
        # 世纪前沿
        adj_nav = self.nav_process(all_data, '世纪前沿指数增强24号私募证券投资基金', '新方程大类配置二号私募证券投资基金')
        nav_dict['500指增']['世纪前沿'] = adj_nav
        # 赫富
        adj_nav = self.nav_process(all_data, '赫富500指数增强一号私募基金', '新方程大类配置基金')
        nav_dict['500指增']['赫富'] = adj_nav.loc['20210602':]
        # a = pd.pivot_table(tmp, index='jzrq', columns='khmc', values='jjjz').sort_index()

        "===========================================500================================================================"
        nav_df_500 = pd.concat(nav_dict['500指增'], axis=1).sort_index()
        sql_script = "SELECT JYRQ as TRADEDATE, ZQMC as INDEXNAME, SPJG as TCLOSE from funddb.ZSJY WHERE ZQDM = '{}' " \
                     "and JYRQ >= {} and JYRQ <= {}".format('000905', nav_df_500.index[0], nav_df_500.index[-1])
        index_data = pd.DataFrame(
            hbs.db_data_query('readonly', sql_script, page_size=5000)['data']).set_index('TRADEDATE')['TCLOSE']
        index_nav = index_data.reindex(nav_df_500.index).to_frame('benchmark')

        excess_return = nav_df_500.pct_change().sub(index_nav.pct_change()['benchmark'].squeeze(), axis=0)
        "===========================================1000==============================================================="
        nav_df_1000 = pd.concat(nav_dict['1000指增'], axis=1).sort_index()
        # 手动添加赫富
        nav_hf = pd.read_excel('D:\\Daily-赫富1000指数增强一号日数据-20220722.xlsx', header=3)
        nav_hf.rename(columns={"净值日期": "trade_date", "累计单位净值": "赫富"}, inplace=True)
        nav_hf['trade_date'] = nav_hf['trade_date'].apply(lambda x: x.replace('-', '') if type(x) == str else datetime.strftime(x, '%Y%m%d'))
        nav_hf = nav_hf.set_index('trade_date')[['赫富']]
        nav_df_1000 = pd.merge(nav_df_1000, nav_hf, left_index=True, right_index=True, how='left')
        # 添加九坤
        nv_series = get_fund_nav_from_sql('20220101', '20220729', {"九坤": "SCP381"})
        nav_df_1000 = pd.merge(nav_df_1000, nv_series, left_index=True, right_index=True, how='left')

        sql_script = "SELECT JYRQ as TRADEDATE, ZQMC as INDEXNAME, SPJG as TCLOSE from funddb.ZSJY WHERE ZQDM = '{}' " \
                     "and JYRQ >= {} and JYRQ <= {}".format('000852', nav_df_1000.index[0], nav_df_1000.index[-1])
        index_data = pd.DataFrame(
            hbs.db_data_query('readonly', sql_script, page_size=5000)['data']).set_index('TRADEDATE')['TCLOSE']
        index_nav = index_data.reindex(nav_df_1000.index).to_frame('benchmark')

        excess_return = nav_df_1000.pct_change().sub(index_nav.pct_change()['benchmark'].squeeze(), axis=0)

        attr_start = '20220701'
        # attr_start = '20210630'
        attr_end = '20220729'
        period_data = excess_return[(excess_return.index >= attr_start) & (excess_return.index <= attr_end)]
        period_data = period_data.dropna(axis=1, how='any').sort_index()

        sql_script = "SELECT * FROM st_ashare.r_st_barra_factor_return where " \
                     "TRADE_DATE >= '{}' and TRADE_DATE <= '{}'".format(attr_start, attr_end)
        res = hbs.db_data_query('alluser', sql_script, page_size=50000)
        factor_return = pd.DataFrame(res['data'])[['trade_date', 'factor_name', 'factor_ret']]
        factor_return = pd.pivot_table(
            factor_return, index='trade_date', columns='factor_name', values='factor_ret').sort_index()[style_names]

        # vif检验
        from statsmodels.stats.outliers_influence import variance_inflation_factor
        vif = [variance_inflation_factor(factor_return.values, factor_return.columns.get_loc(i)) for i in
               factor_return.columns]

        exposure_list = []
        attr_list = []
        for col in period_data.columns:
            y = np.array(period_data[col])
            x = sm.add_constant(factor_return)
            model = sm.OLS(y, x).fit()

            style_attr = x.multiply(model.params)
            style_attr['resid'] = model.resid
            style_attr['alpha'] = style_attr['const'] + style_attr['resid']
            style_attr = style_attr[['alpha'] + style_names]

            # cum
            kt = np.log(1 + period_data[col]) / period_data[col]
            r = (1 + period_data[col]).prod() - 1
            k = np.log(1 + r) / r

            cum_attr = style_attr.T.multiply(kt / k).sum(axis=1)

            exposure = model.params.loc[style_names].to_frame(col)
            exposure_list.append(exposure)

            ret_attr = cum_attr.to_frame(col)
            attr_list.append(ret_attr)

        factor_map_dict = {"size": "规模", "beta": "Beta", "momentum": "动量", "resvol": "波动率", "btop": "估值",
                           "earnyield": "盈利", "growth": "成长", "leverage": "杠杆",
                           "sizenl": "非线性规模", "liquidity": "流动性", "alpha": "alpha"}

        exposure_df = pd.concat(exposure_list, axis=1).T
        exposure_df.columns = [factor_map_dict[x] for x in exposure_df.columns]
        attr_df = pd.concat(attr_list, axis=1).T
        attr_df.columns = [factor_map_dict[x] for x in attr_df.columns]
        attr_df['风格收益'] = attr_df[attr_df.columns[1:]].sum(axis=1)
        attr_df = attr_df[attr_df.columns[1:].tolist() + ['alpha']]


if __name__ == '__main__':
    AlphaNavAttribution('20201231', '20220819', '0549a0f6c1ed408b89aeae920353dea4')




import hbshare as hbs
import pandas as pd
from datetime import datetime
from hbshare.quant.CChen.func import generate_table
import pymysql

pymysql.install_as_MySQLdb()


list_sql_l = '''
    (
    code varchar(10) not null,
    name  varchar(100),
    manager  varchar(100),
    type  varchar(100),
    leverage  int,
    after_fee  int,
    fof  int,
    sales  int,
    first_date  date,
    latest_date  date,
    class  varchar(100),
    benchmark  varchar(100),
    stopped  int,
    primary key (id)
    )
    '''

fund_nw_sql_l = '''
    (
    t_date  date not null,
    nav  float null,
    source  varchar(100) null,
    primary key (t_date)
    )
    '''

funds_table_sql_l = '''
    (
    id bigint not null AUTO_INCREMENT,
    code  varchar(100),
    name  varchar(100),
    t_date  date not null,
    dwjz  float null,
    ljjz  float null,
    nav  float null,
    source  varchar(100) null,
    create_time  datetime,
    primary key (id)
    )
    '''


def update_fund_list_db(funds, db_path, sql_info):

    table = 'fund_list'

    generate_table(
        database='work',
        table=table,
        generate_sql=list_sql_l,
        sql_ip=sql_info['ip'],
        sql_user=sql_info['user'],
        sql_pass=sql_info['pass']
    )
    print(table + ' generated')

    fund_list = pd.DataFrame()
    fund_list = fund_list.append(funds, sort=False).reset_index(drop=True)

    existing_fund = pd.read_sql_query(
        'select * from ' + table, db_path
    )

    for f in range(len(fund_list)):
        print(fund_list['name'][f])
        if fund_list['code'][f] in existing_fund['code'].tolist():
            print(u'\t已存在')
        else:
            fund_list[f: f + 1].to_sql(table, db_path, if_exists='append', index=False)
            print(u'\t录入')


def update_fund_list_db2(db_path, sql_info, data_table='fund_data'):
    # 更新基金净值头尾日期
    table = 'fund_list'

    existing_fund = pd.read_sql_query(
        'select * from ' + table, db_path
    )

    conn = pymysql.connect(
        host=sql_info['ip'], port=3306, database='work',
        user=sql_info['user'], password=sql_info['pass'], charset='utf8'
    )
    cs1 = conn.cursor()

    for f in range(len(existing_fund)):
        print(existing_fund['name'][f])
        fund_data = pd.read_sql_query(
            'select * from ' + data_table
            + ' where code="' + existing_fund['code'][f].upper() + '" order by t_date',
            db_path
        )
        if len(fund_data) == 0:
            print('\tNo data')
            continue
        fund_date_start = fund_data['t_date'][0]
        fund_date_end = fund_data['t_date'][len(fund_data) - 1]

        query = 'update ' + table + ' set first_date=' + datetime.strftime(fund_date_start, '%Y%m%d') \
                + ', latest_date=' + datetime.strftime(fund_date_end, '%Y%m%d') \
                + ' where code="' + existing_fund['code'][f] + '"'

        cs1.execute(query)
        conn.commit()
        print(u'\t初始净值日期更新为：' + datetime.strftime(fund_date_start, '%Y-%m-%d'))
        print(u'\t最新净值日期更新为：' + datetime.strftime(fund_date_end, '%Y-%m-%d'))

    cs1.close()
    conn.close()


# 从hbshare读取基金信息
def update_fund_nav_db4(
        db_path,
        sql_info,
        table='fund_data',
        list_table='fund_list',
        start_date=datetime(1990, 1, 1).date(),
        token=''
):
    generate_table(
        database='work',
        table=table,
        generate_sql=funds_table_sql_l,
        sql_ip=sql_info['ip'],
        sql_user=sql_info['user'],
        sql_pass=sql_info['pass']
    )
    print(table + ' generated')

    fund_list = pd.read_sql_query(
        'select * from ' + list_table, db_path
    )
    for i in range(len(fund_list)):
        code = fund_list['code'][i]
        name = fund_list['name'][i]
        print(code + ' ' + name)

        data_exists = pd.read_sql_query(
            'select `code`, `name`, `t_date`, `nav`, `dwjz`, `ljjz` from ' + table
            + ' where `code`="' + code + '" order by t_date desc',
            db_path
        )

        try:
            data_new = hbs.get_simu_nav_by_code(
                code=code, start_date=start_date.strftime('%Y%m%d'), size=999
            )[0]
        except ValueError:
            data_new = pd.DataFrame()

        # data_daily = hbs.commonQuery(
        #     'FB_RJZ', jjdm=code, access_token=token, fields=['netValueDate', 'netValue', 'fqdwjz', 'ljjz']
        # ).rename(
        #     columns={
        #         'netValueDate': 'jzrq',
        #         'netValue': 'jjjz',
        #     }
        # )
        # if len(data_daily) > 0:
        #     a = 1
        # data_new = data_new.append(data_daily).drop_duplicates(subset=['jzrq'])

        if len(data_new) == 0:
            print('\tNo new data')
        else:
            data_new = data_new.rename(
                columns={
                    'jjjz': 'dwjz',
                    'ljjz': 'ljjz',
                    'fqdwjz': 'nav'
                }
            )
            data_new['t_date'] = pd.to_datetime(data_new['jzrq'].astype(str)).dt.date
            data_new['code'] = code
            data_new['name'] = name
            data_new = data_new[['code', 'name', 't_date', 'nav', 'dwjz', 'ljjz']]
            data_new = data_exists.append(data_new).append(data_exists)
            data_new = data_new.drop_duplicates(['t_date'], keep=False).sort_values(by='t_date').reset_index(drop=True)
            print(data_new)
            data_new['create_time'] = datetime.now()
            data_new.to_sql(table, db_path, if_exists='append', index=False)


def update_fund_nav_out(data, db_path, table='fund_data'):
    fund_list = pd.read_sql_query(
        'select * from fund_list', db_path
    )
    funds = data.columns.tolist()
    funds.remove('t_date')
    for i in funds:
        print(i)
        code = fund_list[fund_list['name'] == i]['code'].tolist()[0]
        nav = data[['t_date', i]].rename(columns={i: 'nav'})

        existing_fund_data = pd.read_sql_query(
            'select * from ' + table + ' where code="' + code + '" order by t_date', db_path
        )
        nav['code'] = code
        nav['name'] = i
        df = existing_fund_data.append(nav[nav['nav'] >= 0].reset_index(drop=True), sort=False)
        df_new = df.drop_duplicates(['t_date'], keep=False).reset_index(drop=True)
        df_new = df_new.append(existing_fund_data).append(existing_fund_data)
        df_new = df_new.drop_duplicates(['t_date'], keep=False).reset_index(drop=True)
        if len(df_new) == 0:
            print('\tNo new data')
            continue
        else:
            print(df_new)
            df_new['create_time'] = datetime.now()
            df_new['source'] = 'out'
        df_new.to_sql(table, db_path, if_exists='append', index=False)





# -*- coding: utf-8 -*-

def singleton(cls, *args, **kw):
    """
    单例decorator
    :param cls: class
    :param args:
    :param kw:
    :return:
    """
    instances = dict()

    def _singleton():
        if cls not in instances:
            instances[cls] = cls(*args, **kw)
        return instances[cls]

    return _singleton()